#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Dict, List, Optional, Union

from th2_act.act_connector import ActConnector
from th2_act.act_parameters import ActParameters
from th2_act.act_response import ActMultiResponse, ActResponse
from th2_act.cache_processor import CacheProcessor
from th2_act.message_response_monitor import MessageResponseMonitor
from th2_grpc_common.common_pb2 import EventID, Message, RequestStatus

logger = logging.getLogger()


class RequestProcessorUtils:

    def __init__(self,
                 act_conn: ActConnector,
                 act_parameters: ActParameters,
                 cache_processor: CacheProcessor):
        self.act_conn = act_conn
        self.act_name = act_parameters.act_name
        self.context = act_parameters.context
        self.cache_processor = cache_processor

        self.act_event_id = act_conn.store_act_event(request_event_id=act_parameters.request_event_id,
                                                     act_name=act_parameters.act_name,
                                                     description=act_parameters.request_description)
        self.checkpoint = self.act_conn.register_checkpoint(act_event_id=self.act_event_id)

        self.response_monitor = MessageResponseMonitor()

    def send_message(self, message: Message) -> ActResponse:
        if self.context.is_active():
            return self.act_conn.send_message(message=message,
                                              act_event_id=self.act_event_id,
                                              checkpoint=self.checkpoint)

        else:
            warning_text = f'{self.act_name} request cancelled by client'
            logger.warning(warning_text)
            self.act_conn.store_send_request_failed_event(message=message, act_event_id=self.act_event_id)

            return ActResponse(status=RequestStatus.ERROR)

    def wait(self, seconds: Union[int, float]) -> None:
        if self.context.time_remaining() > seconds:
            self.response_monitor.await_sync(seconds)
        else:
            logger.warning('Remaining time of context less than wait time. Cache check starts now.')

    def create_received_messages_root_event(self,
                                            method_name: str,
                                            messages_received: int,
                                            messages_expected: Optional[int] = None) -> EventID:
        if messages_expected in (1, None):
            return self.act_conn.store_received_messages_root_event(act_event_id=self.act_event_id,
                                                                    status=RequestStatus.SUCCESS,
                                                                    event_name=f'Receive {messages_received} message(s)'
                                                                               f' - {method_name}')
        elif messages_received == messages_expected:
            return self.act_conn.store_received_messages_root_event(act_event_id=self.act_event_id,
                                                                    status=RequestStatus.SUCCESS,
                                                                    event_name=f'Receive {messages_received}/'
                                                                               f'{messages_expected} message(s)'
                                                                               f' - {method_name}')
        else:
            return self.act_conn.store_received_messages_root_event(act_event_id=self.act_event_id,
                                                                    status=RequestStatus.ERROR,
                                                                    event_name=f'Receive {messages_received}/'
                                                                               f'{messages_expected} message(s)'
                                                                               f' - {method_name}')

    def create_received_messages_events(self,
                                        status_messages_dict: Dict[int, List[Message]],
                                        root_event_id: EventID,
                                        runtime: float) -> None:
        if status_messages_dict:
            for status, messages in status_messages_dict.items():
                for message in messages:
                    self.act_conn.store_process_received_message_succeed_event(received_message=message,
                                                                               root_event_id=root_event_id,
                                                                               status=status)

        else:
            error_text = f'Message matching the condition was not found in {runtime} s'
            self.act_conn.store_process_received_message_failed_event(text=error_text, root_event_id=root_event_id)

    def create_act_responses_list(self,
                                  status_messages_dict: Dict[int, List[Message]],
                                  root_event_id: EventID) -> List[ActResponse]:
        if not self.context.is_active():
            return [self._request_cancelled_act_response(root_event_id)]

        elif status_messages_dict:
            return [
                ActResponse(message=message, status=status, checkpoint=self.checkpoint)
                for status, messages in status_messages_dict.items()
                for message in messages
            ]

        return [ActResponse(status=RequestStatus.ERROR, checkpoint=self.checkpoint)]

    def create_act_multi_response(self,
                                  status_messages_dict: Dict[int, List[Message]],
                                  root_event_id: EventID) -> ActMultiResponse:
        if not self.context.is_active():
            return self._request_cancelled_act_multi_response(root_event_id)

        elif status_messages_dict:
            for status, messages in status_messages_dict.items():
                return ActMultiResponse(messages=messages,
                                        status=status,
                                        checkpoint=self.checkpoint)

        return ActMultiResponse(status=RequestStatus.ERROR, checkpoint=self.checkpoint)

    def _request_cancelled_act_response(self, root_event_id: EventID) -> ActResponse:
        warning_text = f'{self.act_name} cancelled by client'
        logger.warning(warning_text)
        self.act_conn.store_process_received_message_failed_event(text=warning_text, root_event_id=root_event_id)

        return ActResponse(status=RequestStatus.ERROR)

    def _request_cancelled_act_multi_response(self, root_event_id: EventID) -> ActMultiResponse:
        warning_text = f'{self.act_name} cancelled by client'
        logger.warning(warning_text)
        self.act_conn.store_process_received_message_failed_event(text=warning_text, root_event_id=root_event_id)

        return ActMultiResponse(status=RequestStatus.ERROR)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import defaultdict
from itertools import islice
import logging
from time import time
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from th2_act.listener import ParsedMessageListener
from th2_grpc_common.common_pb2 import Message, MessageBatch

logger = logging.getLogger()


StatusMessagesDict = Dict[int, List[Message]]


class CacheProcessor:
    """Stores a cache of received messages. Contains methods to get messages from the cache, to get info about number
    of messages in the cache and to clear the cache.
    """

    def __init__(self, context: Any, prefilter: Optional[Callable]) -> None:
        self.context = context

        if prefilter:
            self.prefilter = prefilter
            self.message_listener = ParsedMessageListener(self._process_incoming_messages_with_prefilter)
        else:
            self.message_listener = ParsedMessageListener(self._process_incoming_messages)

        self.cache: List[Message] = []
        self.last_received_message_index: Optional[int] = None

    def get_n_matching_within_timeout(self,
                                      message_filters: Dict[Callable, int],
                                      n: int,
                                      check_previous_messages: bool,
                                      timeout: Union[int, float],
                                      start_time: float) -> Tuple[StatusMessagesDict, Optional[int]]:
        """Returns first N received messages that match the message_filters from cache.

        The method will check cache until all N messages are found or until context is no more active
        or until timeout is over.
        """

        status_messages_dict: Dict[int, List[Message]] = defaultdict(list)
        matching_message_index: Optional[int] = None
        start_index = self._get_start_index(check_previous_messages)
        messages_to_receive = n

        while messages_to_receive > 0:
            if self.context.is_active() and time() - start_time < timeout:
                stop_index = len(self.cache)

                status_message_iterator = self.get_filtered_message_iterator(message_filters=message_filters,
                                                                             messages_to_receive=messages_to_receive,
                                                                             start_index=start_index,
                                                                             stop_index=stop_index)

                for index, status, message in status_message_iterator:
                    status_messages_dict[status].append(message)
                    matching_message_index = index
                    messages_to_receive -= 1

                start_index = stop_index
            else:
                break

        self._update_last_received_message_index(index=matching_message_index)

        return status_messages_dict, matching_message_index

    def get_n_matching_immediately(self,
                                   message_filters: Dict[Callable, int],
                                   check_previous_messages: bool,
                                   n: Optional[int] = None) -> Tuple[StatusMessagesDict, Optional[int]]:
        """Returns all received messages that match the message_filter from cache.

        The method will check cache only once.
        """

        status_messages_dict: Dict[int, List[Message]] = defaultdict(list)
        matching_message_index: Optional[int] = None
        start_index = self._get_start_index(check_previous_messages)

        if self.context.is_active():
            status_message_iterator = self.get_filtered_message_iterator(message_filters=message_filters,
                                                                         messages_to_receive=n,
                                                                         start_index=start_index,
                                                                         stop_index=len(self.cache))
            for index, status, message in status_message_iterator:
                status_messages_dict[status].append(message)
                matching_message_index = index

        self._update_last_received_message_index(index=matching_message_index)

        return status_messages_dict, matching_message_index

    def get_filtered_message_iterator(self,
                                      message_filters: Dict[Callable, int],
                                      messages_to_receive: Optional[int],
                                      start_index: int,
                                      stop_index: int) -> Iterator[Tuple[int, int, Message]]:
        return islice(
            (
                (index, status, message)
                for index, message in islice(enumerate(self.cache), start_index, stop_index)
                for condition, status in message_filters.items()
                if check(condition, message)
            ),
            0,
            messages_to_receive
        )

    def get_number_of_received_messages(self,
                                        status_messages_dict: Optional[StatusMessagesDict] = None,
                                        before_index: Optional[int] = None,
                                        after_index: Optional[int] = None) -> int:
        if status_messages_dict:
            return sum(len(message_list) for message_list in status_messages_dict.values())
        if before_index is not None:
            return before_index + 1
        if after_index is not None:
            return len(self.cache) - after_index
        else:
            return 0

    def get_all_before_matching_from_cache(self, index: Optional[int]) -> List[Message]:
        if index is not None:
            return self.cache[0:index + 1]
        else:
            return []

    def get_all_after_matching_from_cache(self, index: Optional[int]) -> List[Message]:
        if index is not None:
            return self.cache[index:len(self.cache)]
        else:
            return []

    def clear_cache(self) -> None:
        self.cache.clear()

    @property
    def cache_size(self) -> int:
        return len(self.cache)

    def _get_start_index(self, check_previous_messages: bool) -> int:
        if check_previous_messages or self.last_received_message_index is None:
            return 0
        else:
            return self.last_received_message_index

    def _update_last_received_message_index(self, index: Optional[int]) -> None:
        if index is not None:
            self.last_received_message_index = index

    def _process_incoming_messages(self, consumer_tag: str, message_batch: MessageBatch) -> None:
        try:
            logger.debug('Received MessageBatch with %i messages' % len(message_batch.messages))
            self.cache.extend(message for message in message_batch.messages)

        except Exception as e:
            logger.error('Could not process incoming messages: %s' % e)

    def _process_incoming_messages_with_prefilter(self, consumer_tag: str, message_batch: MessageBatch) -> None:
        try:
            logger.debug('Received MessageBatch with %i messages' % len(message_batch.messages))
            self.cache.extend(message for message in message_batch.messages if check(self.prefilter, message))

        except Exception as e:
            logger.error('Could not process incoming messages: %s' % e)


def check(condition: Callable[[Message], bool], message: Message) -> bool:
    try:
        return condition(message)
    except KeyError:
        return False



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from time import time
from typing import Any, Callable, Dict, List, Optional, Union

from th2_act.act_connector import ActConnector
from th2_act.act_parameters import ActParameters
from th2_act.act_response import ActMultiResponse, ActResponse
from th2_act.cache_processor import CacheProcessor
from th2_act.request_processor_utils import RequestProcessorUtils
from th2_grpc_common.common_pb2 import Message, RequestStatus


logger = logging.getLogger()


class RequestProcessor:
    """A context manager used for sending requests and filtering messages from system under test.

    Mainly used when implementing gRPC methods in ActHandler class.

    Args:
        act_conn (ActConnector): ActConnector class instance. Generates by Act when gRPC server starts.
        act_parameters (ActParameters): ActParameters class instance. Initialize it with act name, gRPC context,
            request event ID and request description.
        prefilter (lambda or function, optional): Filter for messages that the cache will collect.
            Defaults to None - cache will collect all messages.
    """

    def __init__(self,
                 act_conn: ActConnector,
                 act_parameters: ActParameters,
                 prefilter: Optional[Callable] = None):
        self.act_name = act_parameters.act_name
        self._context = act_parameters.context

        self._cache_processor = CacheProcessor(context=self._context, prefilter=prefilter)
        self._rp_utils = RequestProcessorUtils(act_conn=act_conn,
                                               act_parameters=act_parameters,
                                               cache_processor=self._cache_processor)

        self._subscription_manager = act_conn.subscription_manager
        self._subscription_manager.register(message_listener=self._cache_processor.message_listener)

    def __enter__(self) -> Any:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Optional[bool]:
        try:
            self._subscription_manager.unregister(message_listener=self._cache_processor.message_listener)
            if exc_type:
                logger.error('Exception occurred in RequestProcessor (%s): %s. Traceback: %s'
                             % (self.act_name, exc_val, exc_tb.print_tb))
                return True
            return None
        except Exception as e:
            logger.error('Could not stop subscriber monitor: %s' % e)
            return None

    def send(self, message: Message) -> ActResponse:
        """Send message to conn.

        Args:
            message (Message): Message to send to conn.

        Returns:
            ActResponse class instance with status (SUCCESS or ERROR).
        """

        send_request_act_response = self._rp_utils.send_message(message)

        if send_request_act_response.status == RequestStatus.ERROR:
            self._context.cancel()

        return send_request_act_response

    def receive_first_matching(self,
                               message_filters: Dict[Callable, int],
                               timeout: Optional[Union[int, float]] = None,
                               check_previous_messages: bool = True) -> ActResponse:
        """Returns the first message (wrapped in ActResponse class) from cache that matches one of the conditions.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the message.
                The message will be returned as soon as it is found, without waiting for the end of the timeout.
                If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            ActResponse class instance with received message, status and checkpoint.
        """

        start_time = time()

        if timeout is None:
            timeout = self._context.time_remaining()

        status_messages_dict, _ = self._cache_processor.get_n_matching_within_timeout(
            message_filters=message_filters,
            n=1,
            check_previous_messages=check_previous_messages,
            timeout=timeout,
            start_time=start_time
        )

        received_messages_root_event_id = self._rp_utils.create_received_messages_root_event(
            'first matching',
            messages_received=self._cache_processor.get_number_of_received_messages(
                status_messages_dict=status_messages_dict
            ),
            messages_expected=1
        )

        runtime = time() - start_time

        self._rp_utils.create_received_messages_events(status_messages_dict=status_messages_dict,
                                                       root_event_id=received_messages_root_event_id,
                                                       runtime=runtime)

        return self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict,
                                                        root_event_id=received_messages_root_event_id)[0]

    def receive_first_n_matching(self,
                                 message_filters: Dict[Callable, int],
                                 n: int,
                                 timeout: Optional[Union[int, float]] = None,
                                 check_previous_messages: bool = True) -> List[ActResponse]:
        """Returns a list of first N messages (each wrapped in ActResponse class) from cache that match
        the conditions.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            n (int): A number of filtered messages you want to receive.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the messages.
                The messages will be returned as soon as N messages are found, without waiting for the end of the
                timeout. If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint.
        """

        start_time = time()

        if timeout is None:
            timeout = self._context.time_remaining()

        status_messages_dict, _ = self._cache_processor.get_n_matching_within_timeout(
            message_filters=message_filters,
            n=n,
            check_previous_messages=check_previous_messages,
            timeout=timeout,
            start_time=start_time
        )

        received_messages_root_event_id = self._rp_utils.create_received_messages_root_event(
            'first n matching',
            messages_received=self._cache_processor.get_number_of_received_messages(
                status_messages_dict=status_messages_dict
            ),
            messages_expected=n
        )

        runtime = time() - start_time

        self._rp_utils.create_received_messages_events(status_messages_dict=status_messages_dict,
                                                       root_event_id=received_messages_root_event_id,
                                                       runtime=runtime)

        return self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict,
                                                        root_event_id=received_messages_root_event_id)

    def receive_all_before_matching(self,
                                    message_filters: Dict[Callable, int],
                                    timeout: Optional[Union[int, float]] = None,
                                    check_previous_messages: bool = True) -> ActMultiResponse:
        """Returns a list of messages (each wrapped in ActResponse class) from cache that precede a message that
        matches one of the conditions. Message that matching condition is included in the list.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the messages.
                The messages will be returned as soon as the first matching message is found,
                without waiting for the end of the timeout. If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the messages throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint.
        """

        start_time = time()

        if timeout is None:
            timeout = self._context.time_remaining()

        status_messages_dict, index = self._cache_processor.get_n_matching_within_timeout(
            message_filters=message_filters,
            n=1,
            check_previous_messages=check_previous_messages,
            timeout=timeout,
            start_time=start_time
        )

        received_messages_root_event_id = self._rp_utils.create_received_messages_root_event(
            'all before matching',
            messages_received=self._cache_processor.get_number_of_received_messages(before_index=index)
        )

        for status in status_messages_dict:
            status_messages_dict[status] = self._cache_processor.get_all_before_matching_from_cache(index=index)

        runtime = time() - start_time

        self._rp_utils.create_received_messages_events(status_messages_dict=status_messages_dict,
                                                       root_event_id=received_messages_root_event_id,
                                                       runtime=runtime)

        return self._rp_utils.create_act_multi_response(status_messages_dict=status_messages_dict,
                                                        root_event_id=received_messages_root_event_id)

    def receive_all_after_matching(self,
                                   message_filters: Dict[Callable, int],
                                   wait_time: Optional[Union[int, float]] = None,
                                   check_previous_messages: bool = True) -> ActMultiResponse:
        """Returns a list of messages (each wrapped in ActResponse class) from cache that follows a message that
        matches one of the conditions. Message that matching condition is included in the list.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            wait_time (int, optional): The time (in seconds) that the Act will wait before checking the cache for
                the matching message. Defaults to None (cache will be checked immediately and only once).
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint.
       """

        start_time = time()

        if wait_time:
            self._rp_utils.wait(wait_time)

        status_messages_dict, index = self._cache_processor.get_n_matching_immediately(
            message_filters=message_filters,
            check_previous_messages=check_previous_messages,
            n=1
        )

        received_messages_root_event_id = self._rp_utils.create_received_messages_root_event(
            'all after matching',
            messages_received=self._cache_processor.get_number_of_received_messages(after_index=index)
        )

        for status in status_messages_dict:
            status_messages_dict[status] = self._cache_processor.get_all_after_matching_from_cache(index=index)

        runtime = time() - start_time

        self._rp_utils.create_received_messages_events(status_messages_dict=status_messages_dict,
                                                       root_event_id=received_messages_root_event_id,
                                                       runtime=runtime)

        return self._rp_utils.create_act_multi_response(status_messages_dict=status_messages_dict,
                                                        root_event_id=received_messages_root_event_id)

    def receive_all_matching(self,
                             message_filters: Dict[Callable, int],
                             wait_time: Optional[Union[int, float]] = None,
                             check_previous_messages: bool = True) -> List[ActResponse]:
        """Returns a list of messages (wrapped in ActResponse class) from cache that matches conditions.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            wait_time (int, optional): The time that the Act will wait before checking the cache for matching messages.
                Defaults to None (cache will be checked immediately).
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint.
        """

        start_time = time()

        if wait_time:
            self._rp_utils.wait(wait_time)

        status_messages_dict, index = self._cache_processor.get_n_matching_immediately(
            message_filters=message_filters,
            check_previous_messages=check_previous_messages
        )

        received_messages_root_event_id = self._rp_utils.create_received_messages_root_event(
            'all matching',
            messages_received=self._cache_processor.get_number_of_received_messages(
                status_messages_dict=status_messages_dict
            )
        )

        runtime = time() - start_time

        self._rp_utils.create_received_messages_events(status_messages_dict=status_messages_dict,
                                                       root_event_id=received_messages_root_event_id,
                                                       runtime=runtime)

        return self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict,
                                                        root_event_id=received_messages_root_event_id)

    @property
    def cache_size(self) -> int:
        """Returns a number of messages in cache."""

        return self._cache_processor.cache_size

    def clear_cache(self) -> None:
        """Deletes all messages from cache."""

        self._cache_processor.clear_cache()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

import th2_act.act_events as events
from th2_act.act_response import ActResponse
from th2_act.subscription_manager import SubscriptionManager
from th2_common.schema.message.message_router import MessageRouter
from th2_grpc_check1.check1_pb2 import CheckpointRequest, CheckpointResponse
from th2_grpc_check1.check1_service import Check1Service
from th2_grpc_common.common_pb2 import Checkpoint, Event, EventBatch, EventID, Message, MessageBatch, RequestStatus


logger = logging.getLogger()


class ActConnector:
    """Contains methods to send messages and events."""

    def __init__(self,
                 check1_connector: Check1Service,
                 message_router: MessageRouter,
                 event_router: MessageRouter,
                 subscription_manager: SubscriptionManager):
        self.check1_connector = check1_connector
        self.message_router = message_router
        self.event_router = event_router
        self.subscription_manager = subscription_manager

    def store_act_event(self, request_event_id: EventID, act_name: str, description: str) -> EventID:
        event = events.create_act_event(request_event_id=request_event_id,
                                        act_name=act_name,
                                        description=description)
        self._create_and_send_event_batch(event)

        return event.id

    def store_received_messages_root_event(self, act_event_id: EventID, status: int, event_name: str) -> EventID:
        event = events.create_received_messages_root_event(act_event_id=act_event_id,
                                                           status=status,
                                                           event_name=event_name)
        self._create_and_send_event_batch(event)

        return event.id

    def register_checkpoint(self, act_event_id: EventID) -> Checkpoint:
        checkpoint = CheckpointRequest(parent_event_id=act_event_id)
        check1_response: CheckpointResponse = self.check1_connector.createCheckpoint(request=checkpoint)

        return check1_response.checkpoint

    def send_message(self, message: Message, act_event_id: EventID, checkpoint: Checkpoint) -> ActResponse:
        try:
            message.parent_event_id.CopyFrom(act_event_id)

            message_batch = MessageBatch(messages=[message])
            self.message_router.send(message_batch)

            self.store_send_request_succeed_event(message=message, act_event_id=act_event_id)

            return ActResponse(status=RequestStatus.SUCCESS,
                               checkpoint=checkpoint)

        except Exception as e:
            logger.error('Could not send the message with session alias %s: %s'
                         % (message.metadata.id.connection_id.session_alias, e))
            self.store_send_request_failed_event(message=message, act_event_id=act_event_id)

            return ActResponse(status=RequestStatus.ERROR)

    def store_send_request_succeed_event(self, message: Message, act_event_id: EventID) -> EventID:
        event = events.create_send_request_succeed_event(message=message, act_event_id=act_event_id)
        self._create_and_send_event_batch(event)

        return event.id

    def store_send_request_failed_event(self, message: Message, act_event_id: EventID) -> EventID:
        event = events.create_send_request_failed_event(message=message, act_event_id=act_event_id)
        self._create_and_send_event_batch(event)

        return event.id

    def store_process_received_message_succeed_event(self,
                                                     received_message: Message,
                                                     root_event_id: EventID,
                                                     status: int) -> EventID:
        event = events.create_receive_message_succeed_event(received_message=received_message,
                                                            root_event_id=root_event_id,
                                                            status=status)
        self._create_and_send_event_batch(event)

        return event.id

    def store_process_received_message_failed_event(self, text: str, root_event_id: EventID) -> EventID:
        event = events.create_receive_message_failed_event(text, root_event_id)
        self._create_and_send_event_batch(event)

        return event.id

    def _create_and_send_event_batch(self, event: Event) -> None:
        try:
            event_batch = EventBatch(events=[event])
            self.event_router.send_all(message=event_batch)
        except Exception as e:
            logger.error('Could not send event to estore: %s' % e)
